from datetime import datetime
from decimal import Decimal
from typing import List, Optional

from ninja import Schema


class MoonPermisions(Schema):
    view_public_extractions: bool = False
    view_corp_extractions: bool = False
    view_alliance_extractions: bool = False
    view_limited_future: bool = False
    view_observations: bool = False
    view_rentals: bool = False
    edit_rentals: bool = False
    su: bool = False


class IdName(Schema):
    id: int
    name: str
    cat: Optional[str] = None
    cat_id: Optional[int] = None


class Character(Schema):
    character_name: str
    character_id: int
    corporation_name: str
    corporation_id: int
    alliance_name: Optional[str] = None
    alliance_id: Optional[int] = None


class Corporation(Schema):
    corporation_name: str
    corporation_id: int
    alliance_name: Optional[str] = None
    alliance_id: Optional[int] = None


class OreVolume(Schema):
    type: IdName
    volume: Decimal
    total_volume: Decimal
    value: float


class ExtractionEvent(Schema):
    ObserverName: str
    moon: IdName
    system: str
    constellation: str
    region: str
    extraction_end: datetime
    mined_ore: Optional[List[OreVolume]] = None
    jackpot: bool = False
    total_m3: Decimal
    value: Decimal


class MoonRental(Schema):
    moon: IdName
    system: IdName
    contact: Character
    corporation: Corporation
    price: Decimal
    start_date: datetime


class NewMoonRental(Schema):
    moon_id: int
    contact_id: int
    corporation_id: int
    price: Decimal



from corptools.models import EveLocation

from django.core.management.base import BaseCommand

from moons.models import MiningObservation, MoonFrack


class Command(BaseCommand):
    help = 'Fix null Observation Data'

    def handle(self, *args, **options):
        self.stdout.write("Finding dirty Observation Data!")
        obs = MiningObservation.objects.filter(moon__isnull=True)
        structures = obs.values_list("observing_id", flat=True).distinct()
        self.stdout.write(f"Found {structures.count()} structures to fix")
        for s in structures:
            st = EveLocation.objects.get(location_id=s)
            self.stdout.write(f"Checking {st.location_name}")
            mf = MoonFrack.objects.filter(
                structure_id=s, moon_name__isnull=False)
            if mf.exists():
                moon = mf.last().moon_name
                self.stdout.write(
                    f"    Found {moon} for {st.location_name} updating all observations")
                MiningObservation.objects.filter(
                    moon__isnull=True,
                    observing_id=s
                ).update(
                    moon=moon
                )
                self.stdout.write(f"    {st.location_name} Completed")
            else:
                self.stdout.write(f"    Moon Not Found!")



from corptools.models import MapRegion

from django.core.management.base import BaseCommand, CommandError

from moons.models import InvoiceRecord
from moons.tasks import process_moons_from_esi


class Command(BaseCommand):
    help = 'Print Tax Stats'

    def add_arguments(self, parser):
        parser.add_argument('region', type=str)

    def handle(self, *args, **options):

        self.stdout.write("Looking up region!")
        if MapRegion.objects.filter(name=options['region']).exists():
            self.stdout.write(
                f"Sending task to fetch Moons for {options['region']}")
            process_moons_from_esi.delay(
                [MapRegion.objects.filter(name=options['region']).first().region_id])



from datetime import datetime, timedelta

from corptools.models import (
    CorporationAudit, EveItemType, EveLocation, EveName, MapConstellation,
    MapRegion, MapSystem, MapSystemMoon, Notification,
)
from invoices.models import Invoice

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import (
    ExpressionWrapper, F, FloatField, OuterRef, Subquery,
)
from django.db.models.deletion import DO_NOTHING, SET_NULL
from django.utils import timezone

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from . import app_settings
from .managers import MoonManager

if app_settings.discord_bot_active():
    import aadiscordbot
    from aadiscordbot.tasks import send_message

import copy
import hashlib
import json
import logging

from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Model
from django.forms import ValidationError, model_to_dict

logger = logging.getLogger(__name__)


class MoonFrack(models.Model):
    objects = MoonManager()

    corporation = models.ForeignKey(
        CorporationAudit, on_delete=models.CASCADE, related_name='moon')

    moon_name = models.ForeignKey(
        MapSystemMoon, on_delete=models.SET_NULL, null=True, default=None)
    moon_id = models.IntegerField()

    notification = models.ForeignKey(
        Notification, on_delete=models.SET_NULL, null=True, default=None)

    structure = models.ForeignKey(EveLocation, on_delete=models.CASCADE)

    start_time = models.DateTimeField()
    arrival_time = models.DateTimeField()
    auto_time = models.DateTimeField()

    # frack_pinged = models.BooleanField(default=False)
    # finish_notification = models.ForeignKey(Notification, on_delete=models.SET_NULL, null=True, default=None)
    # frack_notification = models.ForeignKey(Notification, on_delete=models.SET_NULL, null=True, default=None)

    class Meta:
        unique_together = (('arrival_time', 'moon_id'),)

    def __str__(self):
        return "{} - {}".format(self.moon_name.name, self.arrival_time)

    class Meta:
        permissions = (('view_available', 'Can View Configured Public Moons'),
                       ('view_corp', 'Can View Own Corps Moons'),
                       ('view_alliance', 'Can View Own Alliances Moons'),
                       ('view_all', 'Can View All Moons'),
                       ('view_limited_future', 'Can View a configured limited subset of future moons'))


class FrackOre(models.Model):
    frack = models.ForeignKey(
        MoonFrack, on_delete=models.CASCADE, related_name='frack')
    ore = models.ForeignKey(EveItemType, on_delete=models.CASCADE)
    total_m3 = models.DecimalField(max_digits=20, decimal_places=2)


# Corp Mining Observation
class MiningObservation(models.Model):
    ob_pk = models.CharField(max_length=50, primary_key=True)

    observing_corporation = models.ForeignKey(
        CorporationAudit, on_delete=models.SET_NULL, null=True, default=None)

    observing_id = models.BigIntegerField(null=True, default=None, blank=True)
    structure = models.ForeignKey(
        EveLocation, on_delete=models.CASCADE, null=True, default=None, blank=True)
    moon = models.ForeignKey(
        MapSystemMoon, on_delete=models.SET_NULL, null=True, default=None, blank=True)

    character_name = models.ForeignKey(
        EveName, on_delete=models.SET_NULL, null=True, default=None)

    character_id = models.IntegerField()
    last_updated = models.DateTimeField()
    quantity = models.BigIntegerField()
    recorded_corporation_id = models.IntegerField()

    type_name = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None)
    type_id = models.IntegerField()

    @classmethod
    def build_pk(cls, corp_id, observer_id, observed_character_id, ob_date, type_id):
        """
        Helper method to get a unique pk for a specific observation. Usefull for bulk updates of data.
        :param corp_id: Observer Corp.
        :param observer_id: Mining Observer ID.
        :param observed_character_id: Character Observed.
        :param ob_date: DateTime of observation.
        :return: :class:String
        """
        date_str = ob_date.strftime("%Y%m%d")
        return "{}-{}-{}-{}-{}".format(corp_id,
                                       observer_id,
                                       observed_character_id,
                                       date_str,
                                       type_id)

    class Meta:
        indexes = (
            models.Index(fields=['observing_id']),
            models.Index(fields=['last_updated']),
            models.Index(fields=['recorded_corporation_id']),
            models.Index(fields=['type_id']),
            models.Index(fields=['character_id'])
        )

    @classmethod
    def explain_taxes(cls, start, end):
        tax_data = {
            "structures": [],
            "taxes": {}
        }
        observers_taxed = []

        taxes = MiningTax.objects.all().order_by('-rank')

        for tax in taxes:
            if tax.id not in tax_data["taxes"]:
                tax_data["taxes"][tax.id] = {
                    "name": str(tax),
                    "structures": []

                }
            observed = MiningObservation.objects.select_related(
                'structure', 'type_name', 'structure__system', 'structure__system__constellation', 'character_name').all()
            observed = observed.filter(
                last_updated__gte=start).filter(last_updated__lt=end)

            if tax.corp:
                observed = observed.filter(
                    observing_corporation__corporation=tax.corp)
            if tax.region:
                observed = observed.filter(
                    structure__system__constellation__region=tax.region)
            if tax.constellation:
                observed = observed.filter(
                    structure__system__constellation=tax.constellation)
            if tax.system:
                observed = observed.filter(structure__system=tax.system)
            if tax.moon:
                observed = observed.filter(moon=tax.moon)

            observed = observed.exclude(structure_id__in=observers_taxed)

            strucutures = list(observed.values_list(
                "structure__location_name", flat=True).distinct())
            tax_data["taxes"][tax.id]['structures'] += strucutures
            tax_data['structures'] += strucutures
            strucuture_ids = list(observed.values_list(
                "structure_id", flat=True).distinct())
            observers_taxed += strucuture_ids
        tax_data["taxes"] = list(tax_data["taxes"].values())
        return tax_data

    @classmethod
    def tax_moons(cls, start, end):
        # get all tax items and return the tax for the time period.

        player_data = {}
        observerd_ids = []
        observers_taxed = []

        taxes = MiningTax.objects.all().order_by('-rank')

        for tax in taxes:
            goo_only = False
            if tax.use_variable_tax:
                goo_only = tax.tax_rate.ignore_ores_in_refine

            type_price = OrePrice.objects.filter(
                item_id=OuterRef('type_id'), goo_only=goo_only)

            observed = MiningObservation.objects.select_related('structure', 'type_name', 'structure__system', 'structure__system__constellation', 'character_name').all() \
                .annotate(isk_value=ExpressionWrapper(
                    Subquery(type_price.values('price')) * F('quantity'),
                    output_field=FloatField()))

            if tax.use_variable_tax:
                tax_price = OreTax.objects.filter(
                    item_id=OuterRef('type_id'), tax=tax.tax_rate)
                observed = observed.annotate(tax_value=ExpressionWrapper(
                    Subquery(tax_price.values('price')) * F('quantity'),
                    output_field=FloatField()))

            observed = observed.filter(
                last_updated__gte=start).filter(last_updated__lt=end)

            if tax.corp:
                observed = observed.filter(
                    observing_corporation__corporation=tax.corp)
            if tax.region:
                observed = observed.filter(
                    structure__system__constellation__region=tax.region)
            if tax.constellation:
                observed = observed.filter(
                    structure__system__constellation=tax.constellation)
            if tax.system:
                observed = observed.filter(structure__system=tax.system)
            if tax.moon:
                observed = observed.filter(moon=tax.moon)

            rate = float(tax.flat_tax_rate)
            # do the ranks
            observed = observed.exclude(structure__in=observers_taxed)
            # print(observed.query)
            # print(observed.count())

            for i in observed.distinct():
                if i.structure not in observers_taxed:
                    observers_taxed.append(i.structure)
                if i.ob_pk not in observerd_ids:
                    observerd_ids.append(i.ob_pk)
                    if i.character_name.eve_id not in player_data:
                        player_data[i.character_name.eve_id] = {}
                        player_data[i.character_name.eve_id]['ores'] = {}
                        player_data[i.character_name.eve_id]['totals_isk'] = 0
                        player_data[i.character_name.eve_id]['tax_isk'] = 0
                        player_data[i.character_name.eve_id]['character_model'] = i.character_name
                        player_data[i.character_name.eve_id]["seen_at"] = []

                    if i.structure.location_name not in player_data[i.character_name.eve_id]["seen_at"]:
                        player_data[i.character_name.eve_id]["seen_at"].append(
                            i.structure.location_name)

                    player_data[i.character_name.eve_id]['totals_isk'] = player_data[i.character_name.eve_id]['totals_isk'] + i.isk_value

                    if tax.use_variable_tax:
                        player_data[i.character_name.eve_id]['tax_isk'] = player_data[i.character_name.eve_id]['tax_isk'] + i.tax_value
                    else:
                        player_data[i.character_name.eve_id]['tax_isk'] = player_data[i.character_name.eve_id]['tax_isk'] + i.isk_value * rate

                    if i.type_name not in player_data[i.character_name.eve_id]['ores']:
                        player_data[i.character_name.eve_id]['ores'][i.type_name.name] = {
                        }
                        player_data[i.character_name.eve_id]['ores'][i.type_name.name]["type_id"] = i.type_id
                        player_data[i.character_name.eve_id]['ores'][i.type_name.name]["value"] = i.isk_value
                        player_data[i.character_name.eve_id]['ores'][i.type_name.name]["count"] = i.quantity
                    else:
                        player_data[i.character_name.eve_id]['ores'][i.type_name.name]["value"] = player_data[
                            i.character_name.eve_id]['ores'][i.type_name.name]["value"]+i.isk_value
                        player_data[i.character_name.eve_id]['ores'][i.type_name.name]["count"] = player_data[
                            i.character_name.eve_id]['ores'][i.type_name.name]["count"]+i.quantity

        output = {
            'player_data': player_data
        }

        return output


class OreTaxRates(models.Model):
    tag = models.CharField(max_length=500, default="Mining Tax")
    refine_rate = models.DecimalField(
        max_digits=5, decimal_places=2, default=87.5)
    ore_rate = models.DecimalField(max_digits=5, decimal_places=2)  # normal
    ubiquitous_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # ubiq
    common_rate = models.DecimalField(max_digits=5, decimal_places=2)  # comon
    uncommon_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # uncom
    rare_rate = models.DecimalField(max_digits=5, decimal_places=2)  # rare
    exceptional_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # best
    ignore_ores_in_refine = models.BooleanField(default=False)

    def __str__(self):
        try:
            return self.tag
        except:
            return "Mining Tax"


# Market History ( GMetrics )
class OrePrice(models.Model):
    item = models.ForeignKey(
        EveItemType, on_delete=models.DO_NOTHING, related_name='ore_price')
    price = models.DecimalField(max_digits=20, decimal_places=2)
    goo_only = models.BooleanField(default=False)
    last_update = models.DateTimeField(auto_now=True)


# tax rates History
class OreTax(models.Model):
    item = models.ForeignKey(
        EveItemType, on_delete=models.DO_NOTHING, related_name='ore_tax')
    price = models.DecimalField(max_digits=20, decimal_places=2)
    last_update = models.DateTimeField(auto_now=True)
    tax = models.ForeignKey(
        OreTaxRates, on_delete=models.CASCADE, related_name='tax_rate')


class MiningTax(models.Model):
    corp = models.ForeignKey(
        EveCorporationInfo, on_delete=models.CASCADE, related_name='moon_mining_tax', null=True, default=None, blank=True)
    tax_rate = models.ForeignKey(
        OreTaxRates, on_delete=models.CASCADE, null=True, default=None, blank=True)
    use_variable_tax = models.BooleanField(default=False)
    flat_tax_rate = models.DecimalField(
        max_digits=5, decimal_places=2, default=0.0)  # best
    region = models.ForeignKey(
        MapRegion, on_delete=models.CASCADE, related_name='tax_region', null=True, default=None, blank=True)
    constellation = models.ForeignKey(
        MapConstellation, on_delete=models.CASCADE, related_name='tax_constellation', null=True, default=None, blank=True)
    system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name='tax_system', null=True, default=None, blank=True)
    moon = models.ForeignKey(
        MapSystemMoon, on_delete=models.CASCADE, related_name='tax_moon', null=True, default=None, blank=True)
    rank = models.IntegerField(default=0, null=True, blank=True)

    def __str__(self):
        area = "Everywhere"
        if self.region:
            area = f"Region: {self.region.name}"
        elif self.constellation:
            area = f"Constellation: {self.constellation.name}"
        elif self.system:
            area = f"System: {self.system.name}"
        elif self.moon:
            area = f"Moon': {self.moon.name}"

        corp = "Everyone"
        if self.corp:
            corp = self.corp.corporation_name

        # return
        rate = ""
        if self.use_variable_tax:
            rate = "Variable ({})".format(self.tax_rate.tag)
        else:
            rate = "of {}%".format(self.flat_tax_rate*100)
        return "Rank {3}: Mining Tax {0} for all `{1}` structures within, {2}".format(rate, corp, area, self.rank)


# tax rates History
class InvoiceRecord(models.Model):
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    ore_prices = models.TextField()
    tax_dump = models.TextField()
    total_mined = models.DecimalField(max_digits=20, decimal_places=2)
    total_taxed = models.DecimalField(max_digits=20, decimal_places=2)
    base_ref = models.CharField(max_length=72, default="")

    @classmethod
    def sanitize_date(cls, date):
        return datetime(year=date.year,
                        month=date.month,
                        day=date.day,
                        tzinfo=date.tzinfo,
                        hour=0,
                        minute=0,
                        second=0)

    @classmethod
    def get_last_invoice_date(cls):
        try:
            return InvoiceRecord.objects.all().order_by('-end_date').first().end_date
        except (ObjectDoesNotExist, AttributeError):
            return datetime.min

    @classmethod
    def generate_invoice(cls, character, ref, amount, message):
        # generate an invoice and return it
        due = timezone.now() + timedelta(days=14)
        return Invoice(character=character,
                       amount=round(amount, -6),
                       invoice_ref=ref,
                       note=message,
                       due_date=due)

    @classmethod
    def ping_invoice(cls, inv):
        # ping the invoice to the user ( if we know them )
        message = "\nPlease check auth for how to pay your mining Taxes!"
        inv.notify(message, title="Mining Taxes")

    @classmethod
    def generate_message(cls, raw_data):
        # Mining data for the invoice message, characters and stations
        pass

    @classmethod
    def generate_inv_ref(cls, char_id, start, end):
        date_str = "%Y%m%d"
        start_str = start.strftime(date_str)
        end_str = end.strftime(date_str)
        return f"MT{char_id}-{start_str}-{end_str}"

    @classmethod
    def generate_invoice_data(cls):
        start_date = cls.sanitize_date(cls.get_last_invoice_date())
        end_date = cls.sanitize_date(timezone.now())
        taxes = {}

        tax_data = MiningObservation.tax_moons(start_date, end_date)
        p_d = copy.deepcopy(tax_data['player_data'])

        all_ownerships = CharacterOwnership.objects.filter(
            character__character_id__in=tax_data['player_data'].keys())

        for o in all_ownerships:
            if o.user.id not in taxes:
                taxes[o.user.id] = {
                    "user": o.user,
                    "locations": set(),
                    "characters": set(),
                    "total_value": 0,
                    "tax_value": 0,
                    "ref": cls.generate_inv_ref(o.user.id, start_date, end_date)
                }

            tx = p_d.pop(o.character.character_id)

            taxes[o.user.id]['characters'].add(tx.get('character_model').name)
            taxes[o.user.id]['locations'].update(tx.get('seen_at'))
            taxes[o.user.id]['total_value'] += tx.get('totals_isk')
            taxes[o.user.id]['tax_value'] += tx.get('tax_isk')

            del tx

        return {"knowns": taxes,
                "unknowns": p_d,
                "start": start_date,
                "end": end_date}

    @classmethod
    def generate_invoices(cls):
        from moons.helpers import OreHelper
        data = cls.generate_invoice_data()
        total_mined = 0
        total_taxed = 0
        # run known people
        for u, d in data['knowns'].items():
            try:
                ref = d['ref']
                amount = d['tax_value']

                total_mined += d['total_value']
                total_taxed += d['tax_value']

                character = d['user'].profile.main_character

                message = f"Mining Taxes for: {', '.join(d['characters'])}\nAt: {', '.join(d['locations'])}"
                inv = cls.generate_invoice(character, ref, amount, message)
                if inv.amount > 0:
                    inv.save()
                    cls.ping_invoice(inv)
            except KeyError:
                pass  # probably wanna ping admin about it.
            except Exception:
                logger.exception(f"Failed to add invoice for {u}:\n\n{d}")

        for u, d in data['unknowns'].items():
            try:
                ref = cls.generate_inv_ref(
                    d['character_model'].eve_id, data['start'], data['end'])
                amount = d['tax_isk']

                total_mined += d['totals_isk']
                total_taxed += d['tax_isk']

                try:
                    character = EveCharacter.objects.get(
                        character_id=d['character_model'].eve_id)
                except EveCharacter.DoesNotExist:
                    character = EveCharacter.objects.create_character(
                        character_id=d['character_model'].eve_id)

                message = f"Mining Taxes for: {character.character_name} \nAt: {', '.join(d['seen_at'])}"
                inv = cls.generate_invoice(character, ref, amount, message)
                if inv.amount > 0:
                    inv.save()
            except KeyError:
                pass  # probably wanna ping admin about it.
            except Exception:
                logger.exception(f"Failed to add invoice for {u}:\n\n{d}")

        return cls.objects.create(start_date=data['start'],
                                  end_date=data['end'],
                                  tax_dump=json.dumps(
                                      data, cls=ExtendedJsonEncoder),
                                  ore_prices=json.dumps(
                                      OreHelper.get_ore_array_with_value(), cls=ExtendedJsonEncoder),
                                  total_mined=total_mined,
                                  total_taxed=total_taxed,
                                  base_ref=cls.generate_inv_ref(
                                      "[id]",  data['start'],  data['end'])
                                  )


class ExtendedJsonEncoder(DjangoJSONEncoder):

    def default(self, o):

        if isinstance(o, User):
            return {"user_id": o.pk}

        if isinstance(o, Model):
            return model_to_dict(o)

        if isinstance(o, set):
            return list(o)

        return super().default(o)


class MoonRental(models.Model):
    note = models.TextField()
    contact = models.ForeignKey(EveCharacter, on_delete=models.CASCADE)
    corporation = models.ForeignKey(
        EveCorporationInfo, on_delete=models.CASCADE)
    moon = models.ForeignKey(MapSystemMoon, on_delete=models.CASCADE)
    price = models.IntegerField(default=100000000)
    start_date = models.DateTimeField()
    end_date = models.DateTimeField(default=None, null=True, blank=True)
    last_invoice = models.ForeignKey(
        Invoice, on_delete=SET_NULL, default=None, null=True, blank=True)

    @classmethod
    def generate_inv_ref(cls, char_id, inv_date, single=False):
        date_str = "%Y%m"
        if single:
            date_str = "%Y%m%d-%H%M"
        start_str = inv_date.strftime(date_str)
        ref = f"MR{char_id}-{start_str}"
        if single:
            ref = f"{ref}-{single}"
        return ref

    @classmethod
    def ping_invoice(cls, inv):
        # ping the invoice to the user ( if we know them )
        message = f"{inv.note}\n\nPlease check auth for how to pay!\n"
        inv.notify(message, title="Moon Rentals")

    @classmethod
    def generate_invoice(cls, cid, moons, price, due_date, single=False):
        msg = f"Moon Rentals for: {', '.join(moons)}"
        if single:
            msg = f"Partial Month Moon Rental for: {', '.join(moons)}"
            single = hashlib.sha1(
                ','.join(moons).encode("UTF-8")).hexdigest()[:8]
        ref = cls.generate_inv_ref(cid, timezone.now(), single=single)
        return Invoice.objects.create(character_id=cid,
                                      amount=round(price, -6),
                                      invoice_ref=ref,
                                      note=msg,
                                      due_date=due_date)

    @classmethod
    def generate_invoices(cls):
        moon_rentals = cls.objects.filter(end_date__isnull=True).select_related(
            "contact", "moon", "contact__character_ownership__user")
        due = timezone.now() + timedelta(days=14)
        total_known = 0
        total_unknown = 0
        users = {}
        for m in moon_rentals:
            try:
                user_id = m.contact.character_ownership.user.profile.main_character_id
                if user_id not in users:
                    users[user_id] = {"uid": user_id,
                                      "moons": [],
                                      "amount": 0,
                                      "characters": set()}
                users[user_id]["moons"].append(f"{m.moon.name} (Ƶ{m.price:,})")
                users[user_id]["amount"] += m.price
                users[user_id]["characters"].add(m.contact.character_name)
                total_known += m.price
            except Exception:
                cls.generate_invoice(
                    m.contact.id, [m.moon.name], m.price, due).save()
                total_unknown += m.price
        for uid, data in users.items():
            inv = cls.generate_invoice(uid, data['moons'], data['amount'], due)
            inv.save()
            cls.ping_invoice(inv)
        if app_settings.discord_bot_active():
            msg = f"Invoiced Moon Rentals, ${total_known:,} to known Users, and ${total_unknown:,} to un-authed characters."
            send_message(msg, app_settings.get_rental_discord_channel())

        return f"Invoiced Moon Rentals, {total_known} to known Users, and {total_unknown} to unknown Users."



from typing import Dict

from corptools.models import EveItemType, InvTypeMaterials, OreTaxRates

from django.db.models import OuterRef, Subquery
from django.utils import timezone

from .models import MiningObservation, MiningTax, OrePrice, OreTax


class OreHelper:

    rank_ids = {
        1923: "exceptional_rate",
        1922: "rare_rate",
        1921: "uncommon_rate",
        1920: "common_rate",
        1884: "ubiquitous_rate"
    }

    asteroids = 25

    def get_mineral_array():
        return list(
            set(
                InvTypeMaterials.objects.filter(
                    eve_type__group__category_id=OreHelper.asteroids
                ).values_list(
                    'met_type_id',
                    flat=True
                )
            )
        )

    def get_ore_array():
        inv_types = InvTypeMaterials.objects.filter(
            eve_type__group__category_id=OreHelper.asteroids
        ).select_related(
            'eve_type',
            'eve_type__group',
            'met_type'
        )

        ore_infos = {}
        for comp in inv_types:
            if comp.eve_type.pk not in ore_infos:
                rarity = OreHelper.rank_ids[
                    comp.eve_type.group_id
                ] if comp.eve_type.group_id in OreHelper.rank_ids else "ore_rate"
                ore_infos[comp.eve_type.pk] = {
                    "minerals": {},
                    "volume": comp.eve_type.packaged_volume,
                    "model": comp.eve_type,
                    "rarity": rarity,
                    "portion": comp.eve_type.portion_size
                }
            ore_infos[comp.eve_type.pk]['minerals'][comp.met_type.name] = comp.qty

        return ore_infos

    def get_detailed_ore_array():
        inv_types = InvTypeMaterials.objects.filter(
            eve_type__group__category_id=OreHelper.asteroids
        ).select_related(
            'eve_type',
            'eve_type__group',
            'met_type'
        )

        ore_infos = {}
        for comp in inv_types:
            if comp.eve_type.pk not in ore_infos:
                rarity = OreHelper.rank_ids[
                    comp.eve_type.group_id
                ] if comp.eve_type.group_id in OreHelper.rank_ids else "ore_rate"
                ore_infos[comp.eve_type.pk] = {
                    "minerals": {},
                    "volume": comp.eve_type.packaged_volume,
                    "model": comp.eve_type,
                    "rarity": rarity,
                    "portion": comp.eve_type.portion_size
                }
            ore_infos[comp.eve_type.pk]['minerals'][comp.met_type.name] = {
                "grp": comp.met_type.group_id,
                "qty": comp.qty
            }

        return ore_infos

    def get_ore_array_with_value():
        input = OreHelper.get_ore_array()
        for o in OrePrice.objects.all():
            if o.goo_only:
                input[o.item_id]['value_goo'] = o.price
            else:
                input[o.item_id]['value'] = o.price
        return input

    def get_ore_array_with_value_and_taxes():
        input = OreHelper.get_ore_array_with_value()
        for t in OreTax.objects.all().select_related("tax"):
            if "tax" not in input[t.item_id]:
                input[t.item_id]['tax'] = {}
            input[t.item_id]['tax'][t.tax.tag] = t.price
        return input

    def set_prices(price_cache: Dict, price_source="the_forge", goo_only=False):
        ores = OreHelper.get_detailed_ore_array()
        print("Goo", goo_only)
        for ore, minerals in ores.items():
            price = 0
            for mineral, detail in minerals["minerals"].items():
                if detail["grp"] == 18 and goo_only:
                    continue
                else:
                    price = price + (detail["qty"] *
                                     price_cache[mineral][price_source])

            OrePrice.objects.update_or_create(
                item=minerals['model'],
                goo_only=goo_only,
                defaults={
                    "price": price / minerals['portion']
                }
            )


def what_frack_id(fracks: dict, observation: dict):
    """
        takes dict of fracks, and a mining observation dict and returns coresponding frack
    """
    date = observation["last_updated"]
    for id, f in fracks.items():
        if f['structure_id'] == observation['structure']:
            compare_to = f["extraction_end"]
            diff = date - compare_to
            if -2 <= diff.days <= 5:
                return id

    return False


def calculate_split_value(prices):
    """
        Calculate Split price from Fuzz API Data
    """
    buy = float(prices["buy"]["max"])
    sell = float(prices["sell"]["min"])
    return ((buy + sell) / 2)



import calendar
import json
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta

import requests
import yaml
from celery import chain, shared_task
from corptools import providers
from corptools.models import (
    CorporationAudit, EveItemType, EveLocation, EveName, MapSystem,
    MapSystemMoon, Notification,
)
from corptools.task_helpers.corp_helpers import get_corp_token
from corptools.task_helpers.update_tasks import fetch_location_name

from django.utils import timezone

from allianceauth.services.tasks import QueueOnce

from . import app_settings
from .app_settings import PUBLIC_MOON_CORPS
from .helpers import OreHelper, calculate_split_value
from .models import (
    FrackOre, InvoiceRecord, MiningObservation, MoonFrack, MoonRental,
    OrePrice, OreTax, OreTaxRates,
)

logger = logging.getLogger(__name__)


def filetime_to_dt(ft):
    us = (ft - 116444736000000000) // 10
    return datetime(1970, 1, 1) + timedelta(microseconds=us)


@shared_task
def cleanup_all_fracks():
    for f in MoonFrack.objects.all().values_list("structure", flat=True).distinct():
        cleanup_moon_pulls.delay(f)


@shared_task
def cleanup_moon_pulls(observer_id):
    fracks = list(MoonFrack.objects.filter(
        structure_id=observer_id).order_by('arrival_time'))
    removals = []
    for index, frack in enumerate(fracks):
        if index+1 < len(fracks):
            if frack.arrival_time > fracks[index + 1].start_time:
                removals.append(frack)
    cnt = len(removals)
    if cnt:
        for df in removals:
            df.delete()
        return f"Removed {cnt} bad extractions."
    return f"No bad extractions."


@shared_task
def process_moon_pulls():
    logger.debug("Started Mining Pull Sync")
    start_time = timezone.now() - timedelta(days=90)

    notification_type_filter = ['MoonminingExtractionStarted', ]

    # YAML Structure
    #    autoTime: #####
    #    moonID: ####
    #    oreVolumeByType:
    #        45490: 3981232.1291307067
    #        45499: 7115599.461194293
    #        45504: 4589835.07632598
    #    readyTime: #####
    #    solarSystemID: ####
    #    startedBy: ####
    #    startedByLink: <a href="showinfo:1385//###">CCCC</a>
    #    structureID: ####
    #    structureLink: <a href="showinfo:35835//####">CCCC</a>
    #    structureName: CCCCCCCCC
    #    structureTypeID: ####

    notifications = Notification.objects.filter(character__character__corporation_id__in=PUBLIC_MOON_CORPS,
                                                timestamp__gte=start_time,
                                                notification_type__in=notification_type_filter).select_related('notification_text')

    fracks = MoonFrack.objects.filter(start_time__gte=start_time)
    notification_ids = set(fracks.values_list(
        'notification__notification_id', flat=True))
    frack_dict = {}
    for frack in fracks:
        if frack.moon_id not in frack_dict:
            frack_dict[frack.moon_id] = {}
        frack_dict[frack.moon_id][frack.start_time] = frack

    _type_list = []
    _ores = []
    for notification in notifications:
        try:
            if notification.notification_id not in notification_ids:
                notification_ids.add(notification.notification_id)
                notification_data = yaml.load(
                    notification.notification_text.notification_text, Loader=yaml.UnsafeLoader)
                moon_id = notification_data['moonID']
                start_time = notification.timestamp

                new_frack = False

                if moon_id in frack_dict:
                    if start_time not in frack_dict[moon_id]:
                        new_frack = True
                else:
                    new_frack = True

                if new_frack:
                    _structure, _c = EveLocation.objects.update_or_create(location_id=notification_data['structureID'],
                                                                          defaults={
                        "location_name": notification_data['structureName'],
                        "system_id": notification_data['solarSystemID']
                    })
                    _moon, _c = MapSystemMoon.objects.get_or_create_from_esi(
                        moon_id=moon_id)
                    _corp_audit = CorporationAudit.objects.get(
                        corporation__corporation_id=notification.character.character.corporation_id)
                    _frack = MoonFrack.objects.create(corporation=_corp_audit,
                                                      moon_name=_moon,
                                                      moon_id=moon_id,
                                                      notification=notification,
                                                      structure=_structure,
                                                      start_time=start_time,
                                                      arrival_time=filetime_to_dt(
                                                          notification_data['readyTime']),
                                                      auto_time=filetime_to_dt(
                                                          notification_data['autoTime'])
                                                      )

                    for ore, volume in notification_data['oreVolumeByType'].items():
                        if ore not in _type_list:
                            _type_list.append(ore)
                        _ores.append(FrackOre(frack=_frack,
                                              ore_id=ore,
                                              total_m3=volume
                                              ))
        except Exception:
            logger.warning("Failed to process moon", exc_info=1)

    EveItemType.objects.create_bulk_from_esi(_type_list)
    FrackOre.objects.bulk_create(_ores, batch_size=500)

    return "fetched!"


@shared_task
def run_obs_for_all_corps(force=False):
    for c in CorporationAudit.objects.all():
        logger.debug(f"Sending Obs update for {c.corporation}")
        queue_moon_obs.apply_async(
            args=[c.corporation.corporation_id], priority=6)


@shared_task
def queue_moon_obs(corp_id, force=False):
    logger.debug("Started Mining Ob Sync for {}".format(corp_id))

    token = get_corp_token(
        corp_id, ['esi-industry.read_corporation_mining.v1'], ['Accountant', 'Director'])
    obs = providers.esi.client.Industry.get_corporation_corporation_id_mining_observers(corporation_id=corp_id,
                                                                                        token=token.valid_access_token()).results()
    task_queue = []
    for ob in obs:
        task_queue.append(process_moon_obs.si(ob.get('observer_id'), corp_id))
    chain(task_queue).apply_async(priority=8)


@shared_task(bind=True, base=QueueOnce)
def process_moon_obs(self, observer_id, corporation_id):
    logger.debug("Started Mining Ob Sync for {}".format(observer_id))

    token = get_corp_token(corporation_id, [
                           'esi-industry.read_corporation_mining.v1', 'esi-universe.read_structures.v1'], ['Accountant', 'Director'])
    obs = providers.esi.client.Industry.get_corporation_corporation_id_mining_observers_observer_id(corporation_id=corporation_id,
                                                                                                    observer_id=observer_id,
                                                                                                    token=token.valid_access_token()).results()
    eve_names = set(EveName.objects.all().values_list('eve_id', flat=True))
    type_names = set(EveItemType.objects.all(
    ).values_list('type_id', flat=True))
    corp = CorporationAudit.objects.get(
        corporation__corporation_id=corporation_id)

    observer = None
    structure_exists = False
    moon = None

    try:
        observer = EveLocation.objects.get(location_id=observer_id)
        structure_exists = True

    except EveLocation.DoesNotExist:
        _ob = fetch_location_name(
            observer_id, "solar_system", token.character_id)
        if _ob:
            _ob.save()
            observer = _ob
            structure_exists = True

    _moon = MoonFrack.objects.filter(
        structure_id=observer_id, moon_name__isnull=False)
    if _moon.exists():
        moon = _moon.last().moon_name

    ob_pks = set(MiningObservation.objects.filter(
        observing_id=observer_id).values_list('ob_pk', flat=True))
    mining_ob_updates = []
    mining_ob_creates = []
    new_eve_names = []
    new_item_types = []
    new_ob_pks = set()
    for ob in obs:
        pk = MiningObservation.build_pk(corporation_id, observer_id, ob.get(
            'character_id'), ob.get('last_updated'), ob.get('type_id'))

        if ob.get('character_id') not in eve_names:
            new_eve_names.append(ob.get('character_id'))
            eve_names.add(ob.get('character_id'))

        if ob.get('type_id') not in type_names:
            new_item_types.append(ob.get('type_id'))
            type_names.add(ob.get('type_id'))

        _ob = MiningObservation(ob_pk=pk,
                                observing_id=observer_id,
                                observing_corporation=corp,
                                character_name_id=ob.get('character_id'),
                                character_id=ob.get('character_id'),
                                last_updated=ob.get('last_updated'),
                                quantity=ob.get('quantity'),
                                recorded_corporation_id=ob.get(
                                    'recorded_corporation_id'),
                                type_id=ob.get('type_id'),
                                type_name_id=ob.get('type_id'),
                                moon=moon)

        if structure_exists:
            _ob.structure = observer

        if pk in ob_pks:
            mining_ob_updates.append(_ob)
        else:
            if pk in new_ob_pks:
                for _test_ob in mining_ob_creates:
                    if _test_ob.pk == pk:
                        _test_ob.quantity += _ob.quantity
                        break
            else:
                mining_ob_creates.append(_ob)
                new_ob_pks.add(pk)

    EveName.objects.create_bulk_from_esi(new_eve_names)
    EveItemType.objects.create_bulk_from_esi(new_item_types)

    if len(mining_ob_creates) > 0:
        MiningObservation.objects.bulk_create(mining_ob_creates)

    if len(mining_ob_updates) > 0:
        MiningObservation.objects.bulk_update(mining_ob_updates, [
                                              'quantity', 'last_updated', 'structure', 'observing_corporation'])

    msg = f"Corp:{corporation_id} Moon:{observer_id} Updated:{len(mining_ob_updates)} Created:{len(mining_ob_creates)}"

    corp.last_update_observers = timezone.now()
    corp.save()

    logger.debug(msg)
    return msg


@shared_task
def update_ore_prices():
    logger.info(
        f"Pulling Ore values from Jita @`{app_settings.MOONS_ORE_RATE_BUY_SELL}`-`{app_settings.MOONS_ORE_RATE_BUCKET}`")
    mins = OreHelper.get_mineral_array()
    minstr = str(mins.pop())
    for i in mins:
        minstr += f",{i}"
    url = f"https://market.fuzzwork.co.uk/aggregates/?station=60003760&types={minstr}"
    response = requests.get(url)
    price_data = response.json()
    price_cache = {}
    for key, item in price_data.items():
        name, created = EveItemType.objects.get_or_create_from_esi(key)
        name = name.name
        if name not in price_cache:
            price_cache[name] = {}
        if app_settings.MOONS_ORE_RATE_BUY_SELL == "split":
            price_cache[name]['the_forge'] = float(calculate_split_value(item))
        else:
            price_cache[name]['the_forge'] = float(
                item[app_settings.MOONS_ORE_RATE_BUY_SELL][app_settings.MOONS_ORE_RATE_BUCKET]
            )

    OreHelper.set_prices(price_cache)
    OreHelper.set_prices(price_cache, goo_only=True)

    update_tax_prices()

    return json.dumps(price_cache, indent=2)


@shared_task
def update_tax_prices():
    taxes = OreTaxRates.objects.all()
    ores = OreHelper.get_ore_array_with_value()

    for tax in taxes:
        print(tax)
        for id, o in ores.items():
            rate = getattr(tax, o['rarity'])
            _o_price = 0.0
            if tax.ignore_ores_in_refine:
                _o_price = float(o['value_goo'])
            else:
                _o_price = float(o['value'])
            price = _o_price * (float(rate)/100) * \
                (float(tax.refine_rate)/100)
            OreTax.objects.update_or_create(
                item=o['model'],
                tax=tax,
                defaults={
                    "price": price
                }
            )


@shared_task
def generate_taxes():
    taxes = InvoiceRecord.generate_invoices()
    return f"Taxes Generated {taxes.base_ref} Total Mined:{taxes.total_mined:,} Total Tax:{taxes.total_taxed:,}"


def _get_system_planet_moons(system_id):
    system = providers.esi.client.Universe.get_universe_systems_system_id(
        system_id=system_id).result()
    out = []
    if system.get('planets', False):
        for p in system.get('planets'):
            if p.get('moons'):
                out += p.get('moons', [])
    return out


@shared_task
def process_moons_from_esi(regions):

    _moons = []
    _moon_models_updates = []
    _moon_models_creates = []

    _processes = []
    _current_systems = MapSystem.objects.filter(
        constellation__region_id__in=regions).values_list('system_id', flat=True)

    providers.esi.client.Status.get_status().result()

    with ThreadPoolExecutor(max_workers=50) as executor:
        for system in _current_systems:
            _processes.append(executor.submit(
                _get_system_planet_moons, system))

    for task in as_completed(_processes):
        _moons += task.result()

    _processes = []
    with ThreadPoolExecutor(max_workers=50) as executor:
        for moon in _moons:
            _processes.append(executor.submit(
                providers.esi._get_moon, moon, False))

    _moon_ids = set(
        list(MapSystemMoon.objects.all().values_list('moon_id', flat=True)))
    for task in as_completed(_processes):
        _moon_ob = task.result()
        if _moon_ob.moon_id in _moon_ids:
            _moon_models_updates.append(_moon_ob)
        else:
            _moon_models_creates.append(_moon_ob)

    if len(_moon_models_creates) > 0:
        MapSystemMoon.objects.bulk_create(
            _moon_models_creates, batch_size=1000)

    if len(_moon_models_updates) > 0:
        MapSystemMoon.objects.bulk_update(
            _moon_models_updates, ['name', 'x', 'y', 'z'], batch_size=1000)

    # sample_mem()
    output = "Moons: (Updated:{}, Created:{}) " \
             .format(len(_moon_models_updates),
                     len(_moon_models_creates))
    # memdump()
    return output


@shared_task
def invoice_moons():
    MoonRental.generate_invoices()


@shared_task
def invoice_single_moon(mrid):
    rental = MoonRental.objects.get(id=mrid)
    rng = calendar.monthrange(rental.start_date.year, rental.start_date.month)
    last_day = datetime(rental.start_date.year,
                        rental.start_date.month, rng[1]).day
    partial_price = round(rental.price / last_day *
                          (last_day-rental.start_date.day), -6)
    if partial_price > 30000000:
        due = timezone.now() + timedelta(days=14)
        inv = MoonRental.generate_invoice(
            rental.contact.id, [rental.moon.name], partial_price, due, single=True)
        inv.save()
        MoonRental.ping_invoice(inv)



from datetime import timedelta

from corptools.models import CorporationAudit, EveLocation

from django.contrib.auth.models import Permission
from django.test import TestCase
from django.utils import timezone

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import (
    EveAllianceInfo, EveCharacter, EveCorporationInfo,
)
from allianceauth.tests.auth_utils import AuthUtils

from ..models import MoonFrack


class TestInvoicesAccessPerms(TestCase):
    def setUp(cls):
        cls.user1 = AuthUtils.create_user('Main User1')
        cls.user2 = AuthUtils.create_user('Main User2')
        cls.user3 = AuthUtils.create_user('Main User3')
        cls.user4 = AuthUtils.create_user('Main User4')
        cls.user5 = AuthUtils.create_user('Main User5')

        corp1 = EveCorporationInfo.objects.create(corporation_id=123,
                                                  corporation_name='corporation.name1',
                                                  corporation_ticker='ABC',
                                                  member_count=1,
                                                  ceo_id=1,
                                                  )

        char1 = EveCharacter.objects.create(character_id=1,
                                            character_name='character.name1',
                                            corporation_id=123,
                                            corporation_name='corporation.name1',
                                            corporation_ticker='ABC')
        char2 = EveCharacter.objects.create(character_id=2,
                                            character_name='character.name2',
                                            corporation_id=123,
                                            corporation_name='corporation.name1',
                                            corporation_ticker='ABC')

        alli1 = EveAllianceInfo.objects.create(alliance_id=3,
                                               alliance_name="test alliance",
                                               alliance_ticker="TEST",
                                               executor_corp_id=2
                                               )
        corp2 = EveCorporationInfo.objects.create(corporation_id=2,
                                                  corporation_name='corporation.name2',
                                                  corporation_ticker='ABC2',
                                                  member_count=1,
                                                  ceo_id=1,
                                                  alliance=alli1
                                                  )

        char3 = EveCharacter.objects.create(character_id=3,
                                            character_name='character.name3',
                                            corporation_id=2,
                                            corporation_name='corporation.name2',
                                            corporation_ticker='ABC2',
                                            alliance_id=3,
                                            alliance_name='test alliance',
                                            alliance_ticker='TEST')
        char4 = EveCharacter.objects.create(character_id=4,
                                            character_name='character.name4',
                                            corporation_id=2,
                                            corporation_name='corporation.name2',
                                            corporation_ticker='ABC',
                                            alliance_id=4,
                                            alliance_name='test alliance2',
                                            alliance_ticker='TEST')

        alli2 = EveAllianceInfo.objects.create(alliance_id=4,
                                               alliance_name="test alliance2",
                                               alliance_ticker="TEST2",
                                               executor_corp_id=3
                                               )
        corp3 = EveCorporationInfo.objects.create(corporation_id=3,
                                                  corporation_name='corporation.name3',
                                                  corporation_ticker='ABC3',
                                                  member_count=1,
                                                  ceo_id=1,
                                                  alliance=alli2
                                                  )

        char5 = EveCharacter.objects.create(character_id=5,
                                            character_name='character.name5',
                                            corporation_id=3,
                                            corporation_name='corporation.name3',
                                            corporation_ticker='ABC3',
                                            alliance_id=4,
                                            alliance_name='test alliance2',
                                            alliance_ticker='TEST')
        char6 = EveCharacter.objects.create(character_id=6,
                                            character_name='character.name6',
                                            corporation_id=3,
                                            corporation_name='corporation.name3',
                                            corporation_ticker='ABC3',
                                            alliance_id=4,
                                            alliance_name='test alliance2',
                                            alliance_ticker='TEST2')

        corp4 = EveCorporationInfo.objects.create(corporation_id=4,
                                                  corporation_name='corporation.name4',
                                                  corporation_ticker='ABC4',
                                                  member_count=1,
                                                  ceo_id=1,
                                                  alliance=alli2
                                                  )

        char7 = EveCharacter.objects.create(character_id=7,
                                            character_name='character.name7',
                                            corporation_id=4,
                                            corporation_name='corporation.name4',
                                            corporation_ticker='ABC4',
                                            alliance_id=4,
                                            alliance_name='test alliance2',
                                            alliance_ticker='TEST2')
        char9 = EveCharacter.objects.create(character_id=9,
                                            character_name='character.name9',
                                            corporation_id=4,
                                            corporation_name='corporation.name4',
                                            corporation_ticker='ABC4',
                                            alliance_id=4,
                                            alliance_name='test alliance2',
                                            alliance_ticker='TEST2'
                                            )

        alli3 = EveAllianceInfo.objects.create(alliance_id=5,
                                               alliance_name="test alliance3",
                                               alliance_ticker="TEST3",
                                               executor_corp_id=4
                                               )
        corp5 = EveCorporationInfo.objects.create(corporation_id=5,
                                                  corporation_name='corporation.name5',
                                                  corporation_ticker='ABC5',
                                                  member_count=1,
                                                  ceo_id=1,
                                                  alliance=alli3
                                                  )
        char8 = EveCharacter.objects.create(character_id=8,
                                            character_name='character.name8',
                                            corporation_id=5,
                                            corporation_name='corporation.name5',
                                            corporation_ticker='ABC5',
                                            alliance_id=5,
                                            alliance_name='test alliance3',
                                            alliance_ticker='TEST3')

        # User 1 - Char 1 and 2 ( Corp 1 no alliance )
        cls.user1.profile.main_character = char1
        CharacterOwnership.objects.create(
            user=cls.user1, character=char1, owner_hash="abc123")
        cls.user1.profile.save()
        CharacterOwnership.objects.create(
            user=cls.user1, character=char2, owner_hash="cba123")
        cls.user1.profile.refresh_from_db()

        # User 2 - Char 3 ( Corp 2, Alliance 1 )
        cls.user2.profile.main_character = char3
        CharacterOwnership.objects.create(
            user=cls.user2, character=char3, owner_hash="cba321")
        cls.user2.profile.save()
        cls.user2.profile.refresh_from_db()

        # User 3 - Char 5 and 7 ( Corp 3 and 4, Alliance 2 )
        cls.user3.profile.main_character = char5
        CharacterOwnership.objects.create(
            user=cls.user3, character=char5, owner_hash="abc432")
        cls.user3.profile.save()
        CharacterOwnership.objects.create(
            user=cls.user3, character=char7, owner_hash="def432")
        cls.user3.profile.refresh_from_db()

        # User 4 - Char 9 and 8 ( Corp 4, 5 Alliance 2, 3)
        cls.user4.profile.main_character = char9
        CharacterOwnership.objects.create(
            user=cls.user4, character=char9, owner_hash="tre456")
        CharacterOwnership.objects.create(
            user=cls.user4, character=char8, owner_hash="wer456")
        cls.user4.profile.save()
        cls.user3.profile.refresh_from_db()

        # User 5 - no characters

        ca1 = CorporationAudit.objects.create(
            corporation=corp1
        )
        ca2 = CorporationAudit.objects.create(
            corporation=corp2
        )
        ca3 = CorporationAudit.objects.create(
            corporation=corp3
        )
        ca4 = CorporationAudit.objects.create(
            corporation=corp4
        )

        cls.view_available_permission = Permission.objects.get_by_natural_key(
            'view_available', 'moons', 'moonfrack')
        cls.view_corp_permission = Permission.objects.get_by_natural_key(
            'view_corp', 'moons', 'moonfrack')
        cls.view_alliance_permission = Permission.objects.get_by_natural_key(
            'view_alliance', 'moons', 'moonfrack')
        cls.view_all_permission = Permission.objects.get_by_natural_key(
            'view_all', 'moons', 'moonfrack')

        cls.user1.user_permissions.add(cls.view_available_permission)

        el1 = EveLocation.objects.create(
            location_id=1,
            location_name="test1"
        )
        el2 = EveLocation.objects.create(
            location_id=2,
            location_name="test2"
        )
        el3 = EveLocation.objects.create(
            location_id=3,
            location_name="test3"
        )
        el4 = EveLocation.objects.create(
            location_id=4,
            location_name="test4"
        )
        el4 = EveLocation.objects.create(
            location_id=5,
            location_name="test5"
        )

        cls.mf1 = MoonFrack.objects.create(
            corporation=ca1,
            moon_id=123,
            structure=el1,
            start_time=timezone.now()-timedelta(days=30),
            arrival_time=timezone.now()-timedelta(hours=1),
            auto_time=timezone.now()
        )
        cls.mf2 = MoonFrack.objects.create(
            corporation=ca2,
            moon_id=456,
            structure=el2,
            start_time=timezone.now()-timedelta(days=30),
            arrival_time=timezone.now()-timedelta(hours=1),
            auto_time=timezone.now()
        )
        cls.mf3 = MoonFrack.objects.create(
            corporation=ca3,
            moon_id=789,
            structure=el3,
            start_time=timezone.now()-timedelta(days=30),
            arrival_time=timezone.now()-timedelta(hours=1),
            auto_time=timezone.now()
        )
        cls.mf4 = MoonFrack.objects.create(
            corporation=ca4,
            moon_id=101,
            structure=el4,
            start_time=timezone.now()-timedelta(days=30),
            arrival_time=timezone.now()-timedelta(hours=1),
            auto_time=timezone.now()
        )

    def test_get_corp_in_alliance(self):
        self.user2.user_permissions.add(self.view_corp_permission)
        cs = MoonFrack.objects.visible_to(self.user2)
        self.assertNotIn(self.mf1, cs)
        self.assertIn(self.mf2, cs)
        self.assertNotIn(self.mf3, cs)
        self.assertNotIn(self.mf4, cs)

    def test_get_corp_in_alliance_with_alliance(self):
        self.user2.user_permissions.add(self.view_alliance_permission)
        cs = MoonFrack.objects.visible_to(self.user2)
        self.assertNotIn(self.mf1, cs)
        self.assertIn(self.mf2, cs)
        self.assertNotIn(self.mf3, cs)
        self.assertNotIn(self.mf4, cs)

    def test_get_alliance_no_alliance(self):
        self.user1.user_permissions.add(self.view_alliance_permission)
        cs = MoonFrack.objects.visible_to(self.user1)
        self.assertIn(self.mf1, cs)
        self.assertNotIn(self.mf2, cs)
        self.assertNotIn(self.mf3, cs)
        self.assertNotIn(self.mf4, cs)

    def test_get_alliance(self):
        self.user3.user_permissions.add(self.view_alliance_permission)
        cs = MoonFrack.objects.visible_to(self.user3)
        self.assertNotIn(self.mf1, cs)
        self.assertNotIn(self.mf2, cs)
        self.assertIn(self.mf3, cs)
        self.assertIn(self.mf4, cs)

    def test_get_alliance_with_corp(self):
        self.user3.user_permissions.add(self.view_alliance_permission)
        self.user3.user_permissions.add(self.view_corp_permission)
        cs = MoonFrack.objects.visible_to(self.user3)
        self.assertNotIn(self.mf1, cs)
        self.assertNotIn(self.mf2, cs)
        self.assertIn(self.mf3, cs)
        self.assertIn(self.mf4, cs)

    def test_global_perms_u1(self):
        self.user1.user_permissions.add(self.view_all_permission)
        cs = MoonFrack.objects.visible_to(self.user1)
        self.assertIn(self.mf1, cs)
        self.assertIn(self.mf2, cs)
        self.assertIn(self.mf3, cs)
        self.assertIn(self.mf4, cs)

    def test_global_perms_u2(self):
        self.user2.user_permissions.add(self.view_all_permission)
        cs = MoonFrack.objects.visible_to(self.user2)
        self.assertIn(self.mf1, cs)
        self.assertIn(self.mf2, cs)
        self.assertIn(self.mf3, cs)
        self.assertIn(self.mf4, cs)

    def test_global_perms_u3(self):
        self.user3.user_permissions.add(self.view_all_permission)
        cs = MoonFrack.objects.visible_to(self.user3)
        self.assertIn(self.mf1, cs)
        self.assertIn(self.mf2, cs)
        self.assertIn(self.mf3, cs)
        self.assertIn(self.mf4, cs)

    def test_su_perms(self):
        cs = MoonFrack.objects.visible_to(self.user5)
        self.assertNotIn(self.mf1, cs)
        self.assertNotIn(self.mf2, cs)
        self.assertNotIn(self.mf3, cs)
        self.assertNotIn(self.mf4, cs)

        self.user5.is_superuser = True

        cs = MoonFrack.objects.visible_to(self.user5)
        self.assertIn(self.mf1, cs)
        self.assertIn(self.mf2, cs)
        self.assertIn(self.mf3, cs)
        self.assertIn(self.mf4, cs)



from django.apps import AppConfig
from django.core.checks import Error, register

from . import __version__, app_settings


class MoonsConfig(AppConfig):
    name = 'moons'
    label = 'moons'

    verbose_name = f"Moons v{__version__}"


@register()
def check_settings(app_configs, **kwargs):
    errors = []
    if app_settings.MOONS_ORE_RATE_BUCKET not in app_settings._valid_buckets:
        errors.append(Error(
            'Invalid MOONS_ORE_RATE_BUCKET',
            hint='Must be one of "weightedAverage", "max", "min", "stddev", "median", "percentile"',
            id='MOONS_ORE_RATE_BUCKET',
        ))
    if app_settings.MOONS_ORE_RATE_BUY_SELL not in app_settings._valid_buy_sell:
        errors.append(Error(
            'Invalid MOONS_ORE_RATE_BUY_SELL',
            hint="Must be 'buy' or 'sell'",
            id='MOONS_ORE_RATE_BUY_SELL',
        ))
    return errors



from django.apps import apps
from django.conf import settings


def get_pub():
    out = getattr(settings, "PUBLIC_MOON_CORPS", [98609787])
    if isinstance(out, list):
        return out
    elif isinstance(out, int):
        return [out]
    elif isinstance(out, str):
        return [int(out)]


PUBLIC_MOON_CORPS = get_pub()  # trust


def discord_bot_active():
    if apps.is_installed("aadiscordbot"):
        import aadiscordbot as ab
        version = ab.__version__.split(".")
        if int(version[0]) >= 3:
            return True
    return False


def get_rental_discord_channel():
    out = getattr(settings, "MOON_RENTAL_CHANNEL", 373640301375651844)
    if isinstance(out, int):
        return out
    elif isinstance(out, str):
        return int(out)


MOONS_ORE_RATE_BUCKET = getattr(
    settings, "MOONS_ORE_RATE_BUCKET", "weightedAverage")
_valid_buckets = [
    "weightedAverage",
    "max",
    "min",
    "stddev",
    "median",
    "percentile"
]
MOONS_ORE_RATE_BUY_SELL = getattr(settings, "MOONS_ORE_RATE_BUY_SELL", "buy")
_valid_buy_sell = [
    "buy",
    "sell",
    "split"
]

MOONS_ENABLE_RENT_COG = getattr(settings, "MOONS_ENABLE_RENT_COG", True)
MOONS_LIMITED_FUTURE_REGIONS = getattr(
    settings, "MOONS_LIMITED_FUTURE_REGIONS", [])
MOONS_LIMITED_FUTURE_DAYS = getattr(settings, "MOONS_LIMITED_FUTURE_DAYS", 7)



import datetime

from corptools.models import EveLocation

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import PermissionDenied
from django.db.models import Sum
from django.shortcuts import render
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from moons.helpers import OreHelper

from . import __version__
from .api import get_moons_and_extractions
from .models import InvoiceRecord, MiningObservation, MoonFrack


@login_required
def extractions(request):
    if request.user.has_perm('moons.view_all'):
        days_to_hold = timezone.now() - datetime.timedelta(days=3)

        events = MoonFrack.objects.visible_to(request.user)
        future_fracks = events.filter(arrival_time__gt=timezone.now())
        current_fracks = events.filter(
            arrival_time__gte=days_to_hold, arrival_time__lt=timezone.now())

    else:
        raise PermissionDenied(
            'You do not have permission to be here. This has been Logged!')

    context = {
        'events': future_fracks,
        'current_fracks': current_fracks,
    }

    return render(request, 'moons/list.html', context=context)


@login_required
@permission_required("admin")
def observers(request):
    all_obs = MiningObservation.objects.all().values('structure').distinct()
    locations = EveLocation.objects.filter(location_id__in=all_obs)

    context = {
        'observers': locations,
    }

    return render(request, 'moons/observers.html', context=context)


@login_required
@permission_required("moons.view_available")
def react(request):
    context = {
        "version": __version__,
        "app_name": "moons",
        "page_title": "Moons"
    }

    return render(request, 'moons/react_base.html', context=context)


@login_required
def moon_report_use(request):
    if not request.user.has_perm('moons.view_all'):
        raise PermissionDenied("No perms to view")

    _ores = OreHelper.get_ore_array_with_value()
    fracks = get_moons_and_extractions(request, 365*10)

    output = []
    for frack in fracks:
        total_value_available = 0
        total_value = 0
        ores = []
        rank = 0
        for o in frack["mined_ore"]:
            total_value_available += \
                _ores[o["type"]["id"]]["value"] * (o["total_volume"]/10)
            total_value += o["value"]
            ores.append(o["type"]["name"])
            if o["type"]["cat_id"] > rank:
                rank = o["type"]["cat_id"]
        ratio = 0
        if total_value > 0:
            ratio = float(total_value)/float(total_value_available)*100
        _o = {
            "Corporation": frack["CorporationName"],
            "moon": frack["moon"]["name"],
            "end_date": frack["extraction_end"],
            "system": frack["system"],
            "constellation": frack["constellation"],
            "region": frack["region"],
            "total_value": total_value_available,
            "mined_value": total_value,
            "mined_ratio": ratio,
            "ores": ", ".join(ores),
            "rank": OreHelper.rank_ids[rank]
        }
        output.append(_o)

    context = {
        "fracks": output
    }

    return render(
        request,
        'moons/dashboards/moon_use_dash.html',
        context=context
    )



from celery import chain
from corptools.models import MapSystem
from corptools.tasks import (
    process_ores_from_esi, update_or_create_map, update_ore_comp_table,
)
from django_celery_beat.models import (
    CrontabSchedule, IntervalSchedule, PeriodicTask,
)

from django.core.management.base import BaseCommand, CommandError

from moons.models import OreTaxRates
from moons.tasks import update_ore_prices


class Command(BaseCommand):
    help = 'Update or Populate Moon Dependent Data'

    def add_arguments(self, parser):
        parser.add_argument('--inline', action='store_true',
                            help='Run update in Console not via Celery')

    def handle(self, *args, **options):
        self.stdout.write("Creating DB models!")
        if OreTaxRates.objects.all().count() == 0:
            OreTaxRates.objects.create(
                tag="Ore Tax",
                ore_rate=20,
                ubiquitous_rate=20,
                common_rate=20,
                uncommon_rate=20,
                rare_rate=20,
                exceptional_rate=20
            )
        self.stdout.write("Setting up Periodic Tasks!")
        schedule_bi_weekly, _ = IntervalSchedule.objects.get_or_create(every=14,
                                                                       period=IntervalSchedule.DAYS)
        schedule_30_min, _ = CrontabSchedule.objects.get_or_create(minute='30',
                                                                   hour='*',
                                                                   day_of_week='*',
                                                                   day_of_month='*',
                                                                   month_of_year='*',
                                                                   timezone='UTC'
                                                                   )

        schedule_start_of_month, _ = CrontabSchedule.objects.get_or_create(minute='0',
                                                                           hour='0',
                                                                           day_of_week='*',
                                                                           day_of_month='1',
                                                                           month_of_year='*',
                                                                           timezone='UTC'
                                                                           )

        task_obs = PeriodicTask.objects.update_or_create(
            task='moons.tasks.run_obs_for_all_corps',
            defaults={
                'crontab': schedule_30_min,
                'interval': None,
                'solar': None,
                'name': 'Moon Obs Updates',
                'enabled': True
            }
        )
        task_frack = PeriodicTask.objects.update_or_create(
            task='moons.tasks.process_moon_pulls',
            defaults={
                'crontab': schedule_30_min,
                'interval': None,
                'solar': None,
                'name': 'Moon Frack Updates',
                'enabled': True
            }
        )
        task_prices = PeriodicTask.objects.update_or_create(
            task='moons.tasks.update_ore_prices',
            defaults={
                'crontab': schedule_start_of_month,
                'interval': None,
                'solar': None,
                'name': 'Moon Ore Prices',
                'enabled': True
            }
        )
        task_invoice = PeriodicTask.objects.update_or_create(
            task='moons.tasks.generate_taxes',
            defaults={
                'interval': schedule_bi_weekly,
                'crontab': None,
                'solar': None,
                'name': 'Send Moon Invoices',
                'enabled': False
            }
        )

        systems = MapSystem.objects.all().count()
        if options['inline']:
            self.stdout.write("Running Tasks inline this may take some time!")
            if systems <= 8285:
                self.stdout.write("Starting Map Update")
                update_or_create_map()
            else:
                self.stdout.write("No map update needed.")
            self.stdout.write("Starting Asteroid Data Update")
            process_ores_from_esi()
            self.stdout.write("Starting Ore Comp Update")
            update_ore_comp_table()
            self.stdout.write("Starting Ore Price Tasks")
            update_ore_prices()
            self.stdout.write("Done!")
        else:
            que = []
            self.stdout.write("Sending Tasks to celery for processing!")
            if systems <= 8285:
                self.stdout.write("Sending Map Update Task")
                que.append(update_or_create_map.si())
            else:
                self.stdout.write("No map update needed.")
            self.stdout.write("Sending Asteroid Data Task")
            que.append(process_ores_from_esi.si())
            self.stdout.write("Sending Ore Comp Task")
            que.append(update_ore_comp_table.si())
            self.stdout.write("Sending Ore Price Tasks")
            que.append(update_ore_prices.si())
            self.stdout.write("Tasks Queued!")
            chain(que).apply_async(priority=4)



from django.urls import path, re_path

from . import views
from .api import api

app_name = 'moons'

urlpatterns = [
    path('', views.extractions, name='list'),
    re_path(r'^data/', views.observers, name='data'),
    re_path(r'^r/', views.react, name='r'),
    re_path(r'^api/', api.urls),
    re_path(r'^report/usage', views.moon_report_use, name='report_use'),

]



# Generated by Django 4.2.8 on 2023-12-13 09:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('moons', '0017_moonrental'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='moonfrack',
            options={'permissions': (('view_available', 'Can View Configured Public Moons'), ('view_corp', 'Can View Own Corps Moons'), ('view_alliance',
                                     'Can View Own Alliances Moons'), ('view_all', 'Can View All Moons'), ('view_limited_future', 'Can View a configured limited subset of future moons'))},
        ),
    ]



from corptools.models import (
    EveItemCategory, EveItemGroup, EveItemType, InvTypeMaterials,
)

from django.test import TestCase

from moons.helpers import OreHelper
from moons.models import OreTaxRates
from moons.tasks import update_tax_prices


class TestInvoicesAccessPerms(TestCase):
    def setUp(cls):
        EveItemCategory.objects.create(category_id=25, name="Asteroid")

        EveItemGroup.objects.create(
            group_id=1923, category_id=25, name="Exceptional Moon Asteroids")
        EveItemGroup.objects.create(
            group_id=1922, category_id=25, name="Rare Moon Asteroids")
        EveItemGroup.objects.create(
            group_id=1921, category_id=25, name="Uncommon Moon Asteroids")
        EveItemGroup.objects.create(
            group_id=1920, category_id=25, name="Common Moon Asteroids")
        EveItemGroup.objects.create(
            group_id=1884, category_id=25, name="Ubiquitous Moon Asteroids")

        EveItemType.objects.create(
            type_id=1,
            name="Mineral 1",
            description="Mineral 1",
            portion_size=1,
            published=True
        )
        EveItemType.objects.create(
            type_id=2,
            name="Mineral 2",
            description="Mineral 2",
            portion_size=1,
            published=True
        )
        EveItemType.objects.create(
            type_id=3,
            name="Mineral 3",
            description="Mineral 3",
            portion_size=1,
            published=True
        )
        EveItemType.objects.create(
            type_id=4,
            name="Mineral 4",
            description="Mineral 4",
            portion_size=1,
            published=True
        )
        EveItemType.objects.create(
            type_id=5,
            name="Mineral 5",
            description="Mineral 5",
            portion_size=1,
            published=True
        )
        EveItemType.objects.create(
            type_id=6,
            name="Mineral 6",
            description="Mineral 6",
            portion_size=1,
            published=True
        )

        EveItemType.objects.create(
            type_id=46319,
            name="Exceptional",
            description="Test Exceptional Ore",
            group_id=1923,
            portion_size=100,
            published=True,
            volume=1000
        )
        # 40	1	46319	46319	1
        InvTypeMaterials.objects.create(
            type_id=46319,
            eve_type_id=46319,
            material_type_id=1,
            met_type_id=1,
            qty=40
        )
        # 40	2	46319	46319	2
        InvTypeMaterials.objects.create(
            type_id=46319,
            eve_type_id=46319,
            material_type_id=2,
            met_type_id=2,
            qty=40
        )
        # 20	3	46319	46319	3
        InvTypeMaterials.objects.create(
            type_id=46319,
            eve_type_id=46319,
            material_type_id=3,
            met_type_id=3,
            qty=20
        )
        # 44	4	46319	46319	4
        InvTypeMaterials.objects.create(
            type_id=46319,
            eve_type_id=46319,
            material_type_id=4,
            met_type_id=4,
            qty=44
        )

        EveItemType.objects.create(
            type_id=46309,
            name="Rare",
            description="Test Rare Ore",
            group_id=1922,
            portion_size=100,
            published=True,
            volume=5000
        )
        # 30	6	46309	46309	6
        InvTypeMaterials.objects.create(
            type_id=46309,
            eve_type_id=46309,
            material_type_id=6,
            met_type_id=6,
            qty=30
        )
        # 20	5	46309	46309	5
        InvTypeMaterials.objects.create(
            type_id=46309,
            eve_type_id=46309,
            material_type_id=5,
            met_type_id=5,
            qty=20
        )
        # 100	3	46309	46309	3
        InvTypeMaterials.objects.create(
            type_id=46309,
            eve_type_id=46309,
            material_type_id=3,
            met_type_id=3,
            qty=100
        )

        EveItemType.objects.create(
            type_id=46301,
            name="Uncommon",
            description="Test Uncommon Ore",
            group_id=1921,
            portion_size=100,
            published=True,
            volume=5000
        )
        # 45	1	46301	46301	1
        InvTypeMaterials.objects.create(
            type_id=46301,
            eve_type_id=46301,
            material_type_id=1,
            met_type_id=1,
            qty=45
        )
        # 25	2	46301	46301	2
        InvTypeMaterials.objects.create(
            type_id=46301,
            eve_type_id=46301,
            material_type_id=2,
            met_type_id=2,
            qty=25
        )

        EveItemType.objects.create(
            type_id=46295,
            name="Common",
            description="Test Common Ore",
            group_id=1920,
            portion_size=100,
            published=True,
            volume=5000
        )
        # 500	5	46295	46295	5
        InvTypeMaterials.objects.create(
            type_id=46295,
            eve_type_id=46295,
            material_type_id=5,
            met_type_id=5,
            qty=500
        )

        EveItemType.objects.create(
            type_id=46285,
            name="Ubiquitous",
            description="Test Ubiquitous Ore",
            group_id=1884,
            portion_size=100,
            published=True,
            volume=5000
        )
        # 50	1	46285	46285	1
        InvTypeMaterials.objects.create(
            type_id=46285,
            eve_type_id=46285,
            material_type_id=5,
            met_type_id=5,
            qty=50
        )
        # 25	2	46285	46285	2
        InvTypeMaterials.objects.create(
            type_id=46285,
            eve_type_id=46285,
            material_type_id=2,
            met_type_id=2,
            qty=25
        )

        OreTaxRates.objects.create(
            tag="test_50",
            refine_rate=50,
            ore_rate=50,
            ubiquitous_rate=50,
            common_rate=50,
            uncommon_rate=50,
            rare_rate=50,
            exceptional_rate=50,
        )
        OreTaxRates.objects.create(
            tag="test_cascade",
            refine_rate=70,
            ore_rate=10,
            ubiquitous_rate=20,
            common_rate=30,
            uncommon_rate=40,
            rare_rate=50,
            exceptional_rate=60,
        )

    def test_ore_builder(self):
        ores = OreHelper.get_ore_array()
        self.assertEqual(len(ores), 5)

        self.assertEqual(len(ores[46319]['minerals']), 4)
        self.assertEqual(sum(list(ores[46319]['minerals'].values())), 144)

        self.assertEqual(len(ores[46309]['minerals']), 3)
        self.assertEqual(sum(list(ores[46309]['minerals'].values())), 150)

        self.assertEqual(len(ores[46301]['minerals']), 2)
        self.assertEqual(sum(list(ores[46301]['minerals'].values())), 70)

        self.assertEqual(len(ores[46295]['minerals']), 1)
        self.assertEqual(sum(list(ores[46295]['minerals'].values())), 500)

        self.assertEqual(len(ores[46285]['minerals']), 2)
        self.assertEqual(sum(list(ores[46285]['minerals'].values())), 75)

    def test_full_prices(self):
        price_cache = {
            "Mineral 1": {"the_forge": 100},
            "Mineral 2": {"the_forge": 200},
            "Mineral 3": {"the_forge": 300},
            "Mineral 4": {"the_forge": 400},
            "Mineral 5": {"the_forge": 500},
            "Mineral 6": {"the_forge": 1000}
        }
        OreHelper.set_prices(price_cache)
        update_tax_prices()

        ores = OreHelper.get_ore_array_with_value_and_taxes()

        self.assertEqual(len(ores), 5)
        self.assertEqual(float(ores[46319]['value']), 356)
        self.assertEqual(float(ores[46309]['value']), 700)
        self.assertEqual(float(ores[46301]['value']), 95)
        self.assertEqual(float(ores[46295]['value']), 2500)
        self.assertEqual(float(ores[46285]['value']), 300)

        self.assertEqual(float(ores[46319]['tax']['test_50']), 89)
        self.assertEqual(float(ores[46309]['tax']['test_50']), 175)
        self.assertEqual(float(ores[46301]['tax']['test_50']), 23.75)
        self.assertEqual(float(ores[46295]['tax']['test_50']), 625)
        self.assertEqual(float(ores[46285]['tax']['test_50']), 75)

        self.assertEqual(float(ores[46319]['tax']['test_cascade']), 149.52)
        self.assertEqual(float(ores[46309]['tax']['test_cascade']), 245)
        self.assertEqual(float(ores[46301]['tax']['test_cascade']), 26.60)
        self.assertEqual(float(ores[46295]['tax']['test_cascade']), 525)
        self.assertEqual(float(ores[46285]['tax']['test_cascade']), 42)

    def test_minerals(self):
        ores = OreHelper.get_mineral_array()
        self.assertEqual(len(ores), 6)
        self.assertEqual(sum(ores), 21)



# Generated by Django 4.2.13 on 2024-08-22 11:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('moons', '0018_alter_moonfrack_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='oreprice',
            name='goo_only',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='oretaxrates',
            name='ignore_ores_in_refine',
            field=models.BooleanField(default=False),
        ),
    ]



import logging
from datetime import timedelta
from typing import List

from corptools.models import CharacterAudit, CorporationAudit, MapSystemMoon
# from invoices.models import Invoice
from ninja import Form, NinjaAPI
from ninja.security import django_auth

from django.conf import settings
from django.db.models import (
    ExpressionWrapper, F, FloatField, Max, OuterRef, Subquery, Sum,
)
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from esi.models import Token

from moons.helpers import OreHelper, what_frack_id

from . import app_settings, models, schema

logger = logging.getLogger(__name__)


api = NinjaAPI(title="MoonTool API", version="0.0.1",
               urls_namespace='moons:api', auth=django_auth, csrf=True,
               openapi_url=settings.DEBUG and "/openapi.json" or "")


@api.get(
    "/user/permissions",
    response={200: schema.MoonPermisions},
    tags=["User"]
)
def get_user_permissions(request):
    return {
        "view_public_extractions": request.user.has_perm('moons.view_available'),
        "view_corp_extractions": request.user.has_perm('moons.view_corp'),
        "view_alliance_extractions": request.user.has_perm('moons.view_alliance'),
        "view_limited_future": request.user.has_perm('moons.view_limited_future'),
        "view_observations": request.user.has_perm('moons.view_all'),
        "view_rentals": request.user.has_perm('moons.view_moonrental'),
        "edit_rentals": request.user.has_perm('moons.change_moonrental'),
        "su": request.user.is_superuser
    }


JACKPOT_IDS = [
    46281,  # Glistening Zeolites
    46283,  # Glistening Sylvite
    46285,  # Glistening Bitumens
    46287,  # Glistening Coesite
    46305,  # Glowing Carnotite
    46307,  # Glowing Zircon
    46309,  # Glowing Pollucite
    46311,  # Glowing Cinnabar
    46297,  # Shimmering Otavite
    46299,  # Shimmering Sperrylite
    46301,  # Shimmering Vanadinite
    46303,  # Shimmering Chromite
    46289,  # Twinkling Cobaltite
    46291,  # Twinkling Euxenite
    46293,  # Twinkling Titanite
    46295,  # Twinkling Scheelite
    46313,  # Shining Xenotime
    46315,  # Shining Monazite
    46317,  # Shining Loparite
    46319,  # Shining Ytterbite
]


@api.get(
    "/extractions/",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_moons_and_obs(request):
    if not request.user.has_perm("moons.view_available"):
        return []
    past_days = 3

    return get_moons_and_extractions(request, past_days)


@api.get(
    "/extractions/past",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_moons_and_obs_past(request):
    if not request.user.has_perm("moons.view_all"):
        return []

    past_days = 30 * 12  # 12 months = max pull 6 events per moon

    return get_moons_and_extractions(request, past_days)


def get_moons_and_extractions(request, past_days):
    start_date = timezone.now() - timedelta(days=past_days)
    time_from = timezone.now() - timedelta(days=past_days + 1)

    events = models.MoonFrack.objects.visible_to(request.user)
    current_fracks = events.filter(
        arrival_time__gte=start_date,
        arrival_time__lt=timezone.now()
    ).select_related(
        "moon_name",
        "moon_name__system",
        "moon_name__system__constellation",
        "moon_name__system__constellation__region",
    ).prefetch_related(
        'frack',
        "frack__ore",
        "frack__ore__group"
    )

    type_price = models.OrePrice.objects.filter(
        item_id=OuterRef('type_id'),
        goo_only=False
    )

    output = {}
    str_ob_dict = {}

    for e in current_fracks:
        output[e.id] = {
            "CorporationName": e.corporation.corporation.corporation_name,
            "ObserverName": e.structure.location_name,
            "system": e.moon_name.system.name,
            "constellation": e.moon_name.system.constellation.name,
            "region": e.moon_name.system.constellation.region.name,
            "moon": {
                "name": e.moon_name.name,
                "id": e.moon_id
            },
            "extraction_end": e.arrival_time,
            "mined_ore": [],
            "total_m3": 0,
            "value": 0,
            "structure_id": e.structure.location_id
        }

        for o in e.frack.all():
            if e.id not in str_ob_dict:
                str_ob_dict[e.id] = {}
            output[e.id]['total_m3'] += o.total_m3
            str_ob_dict[e.id][o.ore.name] = {
                "type": {
                    "id": o.ore_id,
                    "name": o.ore.name,
                    "cat": o.ore.group.name,
                    "cat_id": o.ore.group_id
                },
                "volume": 0,
                "total_volume": o.total_m3,
                "value": 0
            }

    observations = models.MiningObservation.objects \
        .filter(last_updated__gte=time_from) \
        .filter(observing_id__in=current_fracks.values_list("structure_id", flat=True)) \
        .values('structure', 'type_id', "last_updated") \
        .annotate(mined=(Sum('quantity') * F('type_name__volume'))) \
        .annotate(ore_value=ExpressionWrapper(
            Subquery(type_price.values('price')) * Sum('quantity'),
            output_field=FloatField())) \
        .annotate(name=F('type_name__name'))

    for o in observations:
        frack = what_frack_id(output, o)
        if frack is False:
            continue
        nme = o["name"].split(" ")[-1]
        if frack in str_ob_dict:
            if request.user.has_perm("moons.view_all"):
                str_ob_dict[frack][nme]["value"] += o["ore_value"]
                output[frack]['value'] += o["ore_value"]
            str_ob_dict[frack][nme]["volume"] += o['mined']

        if o['type_id'] in JACKPOT_IDS:
            output[frack]["jackpot"] = True

    for s, o in str_ob_dict.items():
        output[s]["mined_ore"] = list(o.values())

    return list(output.values())


@api.get(
    "/extractions/future",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_future_extractions(request):
    perm_view_all = request.user.has_perm("moons.view_all")
    perm_view_limited = request.user.has_perm("moons.view_limited_future")
    if not (perm_view_all or perm_view_limited):
        return []

    start_date = timezone.now()

    events = models.MoonFrack.objects.visible_to(request.user)

    if perm_view_limited and not perm_view_all:
        end_time = timezone.now() + timedelta(days=app_settings.MOONS_LIMITED_FUTURE_DAYS)
        events = events.filter(
            arrival_time__lte=end_time,
            moon_name__system__constellation__region_id__in=app_settings.MOONS_LIMITED_FUTURE_REGIONS
        )

    current_fracks = events.filter(
        arrival_time__gte=start_date).select_related(
            "moon_name",
            "moon_name__system",
            "moon_name__system__constellation",
            "moon_name__system__constellation__region",
    ).prefetch_related('frack',
                       "frack__ore",
                       "frack__ore__group"
                       )

    type_prices = OreHelper.get_ore_array_with_value()

    output = {}
    str_ob_dict = {}

    for e in current_fracks:
        output[e.structure_id] = {
            "ObserverName": e.structure.location_name,
            "system": e.moon_name.system.name,
            "constellation": e.moon_name.system.constellation.name,
            "region": e.moon_name.system.constellation.region.name,
            "moon": {
                "name": e.moon_name.name,
                "id": e.moon_id
            },
            "extraction_end": e.arrival_time,
            "mined_ore": [],
            "total_m3": 0,
            "value": 0
        }
        for o in e.frack.all():
            value = int(float(
                o.total_m3) / float(type_prices[o.ore_id]["volume"]) * float(type_prices[o.ore_id]["value"]))
            if e.structure_id not in str_ob_dict:
                str_ob_dict[e.structure_id] = {}
            output[e.structure_id]['total_m3'] += o.total_m3
            output[e.structure_id]['value'] += value

            str_ob_dict[e.structure_id][o.ore.name] = {
                "type": {
                    "id": o.ore_id,
                    "name": o.ore.name,
                    "cat": o.ore.group.name,
                    "cat_id": o.ore.group_id
                },
                "volume": 0,
                "total_volume": o.total_m3,
                "value": value
            }

    for s, o in str_ob_dict.items():
        output[s]["mined_ore"] = list(o.values())

    return list(output.values())


@api.get(
    "/moons/search",
    response={200: List[schema.IdName]},
    tags=["Search"]
)
def get_moon_search(request, search_text: str, limit: int = 10):
    return MapSystemMoon.objects.filter(name__icontains=search_text).values("name", id=F("moon_id"))[:limit]


@api.get(
    "/corporations/search",
    response={200: List[schema.Corporation]},
    tags=["Search"]
)
def get_corporation_search(request, search_text: str, limit: int = 10):
    return EveCorporationInfo.objects.filter(corporation_name__icontains=search_text)[:limit]


@api.get(
    "/characters/search",
    response={200: List[schema.Character]},
    tags=["Search"]
)
def get_character_search(request, search_text: str, limit: int = 10):
    return EveCharacter.objects.filter(character_name__icontains=search_text)[:limit]


@api.get(
    "/rental/list",
    response={200: List[schema.MoonRental]},
    tags=["Rentals"]
)
def get_moon_rentals(request):
    if not request.user.has_perm("moons.add_moonrental"):
        return []

    rentals = models.MoonRental.objects.filter(end_date__isnull=True).select_related(
        "moon", "moon__system", "contact", "corporation")
    out = []
    for r in rentals:
        out.append(
            {"moon": {
                "id": r.moon.moon_id,
                "name": r.moon.name
            },
                "system": {
                "id": r.moon.system.system_id,
                "name": r.moon.system.name
            },
                "contact": r.contact,
                "corporation": r.corporation,
                "price": r.price,
                "start_date": r.start_date
            }
        )

    return out


# @api.get(
#     "/rental/payments",
#     response={200: List[schema.MoonRental]},
#     tags=["Rentals"]
# )
# def get_moon_rental_payments(request):
#     if not request.user.has_perm("moons.add_moonrental"):
#         return []

#     payments = Invoice.objects.filter(paid=False).select_related(
#         "moon", "moon__system", "contact", "corporation")
#     out = []
#     for r in rentals:
#         out.append(
#             {"moon": {
#                 "id": r.moon.moon_id,
#                 "name": r.moon.name
#             },
#                 "system": {
#                 "id": r.moon.system.system_id,
#                 "name": r.moon.system.name
#             },
#                 "contact": r.contact,
#                 "corporation": r.corporation,
#                 "price": r.price,
#                 "start_date": r.start_date
#             }
#         )

#     return out


@api.post(
    "/rental/new",
    response={200: schema.MoonRental, 403: str},
    tags=["Rentals"]
)
def post_moon_rental_new(request, rental: schema.NewMoonRental = Form(...)):
    if not request.user.has_perm("moons.add_moonrental"):
        return 403, "Permission Denied!"

    if models.MoonRental.objects.filter(moon_id=rental.moon_id, end_date__isnull=True).exists():
        return 403, "Moon Already Rented!"
    try:
        char = EveCharacter.objects.get(character_id=rental.contact_id)
    except EveCharacter.DoesNotExist:
        return 403, "Character Unknown to Auth"
    try:
        corp = EveCorporationInfo.objects.get(
            corporation_id=rental.corporation_id)
    except EveCorporationInfo.DoesNotExist:
        return 403, "Corporation Unknown to Auth"
    new_rental = models.MoonRental.objects.create(
        moon_id=rental.moon_id,
        contact=char,
        corporation=corp,
        price=rental.price,
        start_date=timezone.now()
    )
    return 200, {"moon": {
        "id": new_rental.moon.moon_id,
        "name": new_rental.moon.name
    },
        "system": {
        "id": new_rental.moon.system.system_id,
        "name": new_rental.moon.system.name
    },
        "contact": new_rental.contact,
        "corporation": new_rental.corporation,
        "price": new_rental.price,
        "start_date": new_rental.start_date
    }


@api.get(
    "/admin/list",
    response={200: List, 403: str},
    tags=["Admin"]
)
def get_corp_stats(request):
    if not request.user.is_superuser:
        return 403, "Permission Denied!"

    corps = CorporationAudit.objects.visible_to(request.user)

    output = []

    for c in corps:
        _c = {
            "name": c.corporation.corporation_name,
            "char_tokens": 0,
            "corp_tokens": 0,
            "obs": c.last_update_observers,
            "frack": "Never"
        }

        chars = CharacterAudit.objects.filter(character__corporation_id=c.corporation.corporation_id,
                                              characterroles__accountant=True,
                                              active=True)

        _c["char_tokens"] = chars.count()
        dt = chars.aggregate(Max("last_update_notif"))
        _c["frack"] = dt["last_update_notif__max"]
        scopes = ['esi-industry.read_corporation_mining.v1',
                  'esi-universe.read_structures.v1']

        tokens = Token.objects \
            .filter(character_id__in=chars.values_list("character__character_id", flat=True)) \
            .require_scopes(scopes)

        _c["corp_tokens"] = tokens.count()

        output.append(_c)

    return output


@api.get(
    "/admin/explain",
    response={200: dict, 403: str},
    tags=["Admin"]
)
def get_explain_tax(request):
    if not request.user.is_superuser:
        return 403, "Permission Denied!"
    last_date = models.InvoiceRecord.get_last_invoice_date()
    data = models.MiningObservation.explain_taxes(last_date, timezone.now())
    return data


@api.get(
    "/admin/outstanding",
    response={200: list, 403: str},
    tags=["Admin"]
)
def get_outstanding_tax(request):
    if not request.user.is_superuser:
        return 403, "Permission Denied!"
    output = []
    last_date = models.InvoiceRecord.get_last_invoice_date()
    output.append(f"Last Invoice {last_date}!")
    locations = set()
    data = models.InvoiceRecord.generate_invoice_data()
    total_mined = 0
    total_taxed = 0
    # run known people
    for u, d in data['knowns'].items():
        try:
            total_mined += d['total_value']
            total_taxed += d['tax_value']
            for loc in d['locations']:
                locations.add(loc)
        except KeyError:
            pass  # probably wanna ping admin about it.

    for u, d in data['unknowns'].items():
        try:
            total_mined += d['totals_isk']
            total_taxed += d['tax_isk']
            for loc in d['seen_at']:
                locations.add(loc)
        except KeyError:
            pass  # probably wanna ping admin about it.

    output.append(
        f"We've seen {len(data['knowns'])} known members!"
    )
    output.append(
        f"We've seen { len(data['unknowns']) } unknown characters!"
    )
    output.append(
        f"Who have mined ${total_mined:,} worth of ore!"
    )
    output.append(
        f"Current Tax puts this at ${total_taxed:,} in taxes!\n"
    )
    output.append(
        "the structures included are:"
    )

    for s in locations:
        output.append(f"  - {s}")

    return output



# Cog Stuff
import logging
# AA Contexts
from typing import Optional

from aadiscordbot.app_settings import get_all_servers
from aadiscordbot.cogs.utils.decorators import has_any_perm
from aadiscordbot.utils.auth import get_auth_user
from corptools.models import CorporationAudit, MapSystemMoon, Structure
from discord import AutocompleteContext, Interaction, option
from discord.commands import SlashCommandGroup
from discord.embeds import Embed
from discord.ext import commands

from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from moons import app_settings
from moons.models import InvoiceRecord, MoonFrack, MoonRental

logger = logging.getLogger(__name__)

BLUE = 0x3498db
MAGENTA = 0xe91e63
GREYPLE = 0x99aab5
RED = 0x992d22


class MoonsCog(commands.Cog):
    """
    All about Moons!
    """

    def __init__(self, bot):
        self.bot = bot

    pinger_commands = SlashCommandGroup(
        "moons", "Moon Module Commands", guild_ids=get_all_servers())

    def sender_has_moon_perm(self, ctx):
        has_perm = has_any_perm(ctx.author.id, ["moons.view_all"], guild=ctx.guild)
        if has_perm:
            return True
        else:
            return False

    def sender_has_corp_moon_perm(self, ctx):
        has_perm = has_any_perm(ctx.author.id, ["moons.view_corp"], guild=ctx.guild)
        if has_perm:
            return True
        else:
            return False

    def sender_has_moon_rental_create_perm(self, ctx):
        has_perm = has_any_perm(ctx.author.id, ["moons.change_moonrental"], guild=ctx.guild)
        if has_perm:
            return True
        else:
            return False

    @pinger_commands.command(name='print_stats', guild_ids=get_all_servers())
    async def info_slash(self, ctx):
        """
        Print the Uninvocied Mining Stats!
        """

        if not self.sender_has_moon_perm(ctx):
            return await ctx.respond(f"You do not have permision to use this command.", ephemeral=True)

        await ctx.respond(f"Calculating the Moon Stats.")

        last_date = InvoiceRecord.get_last_invoice_date()
        date_str = last_date.strftime('%Y/%m/%d')
        e = Embed(title=f"Last Invoice {date_str}")
        accounts_seen = 0
        locations = set()
        data = InvoiceRecord.generate_invoice_data()
        total_mined = 0
        total_taxed = 0
        # run known people
        for u, d in data['knowns'].items():
            try:
                total_mined += d['total_value']
                total_taxed += d['tax_value']

                accounts_seen += 1
                for l in d['locations']:
                    locations.add(l)
            except KeyError:
                pass  # probably wanna ping admin about it.

        for u, d in data['unknowns'].items():
            try:
                total_mined += d['totals_isk']
                total_taxed += d['tax_isk']
                for l in d['seen_at']:
                    locations.add(l)
            except KeyError:
                pass  # probably wanna ping admin about it.

        e.add_field(name="Known Members",
                    value=f"{accounts_seen}", inline=False)
        e.add_field(name="Unknown Characters",
                    value=f"{len(data['unknowns'])}", inline=False)
        e.add_field(name="Total Mined",
                    value=f"${total_mined:,}", inline=False)
        e.add_field(name="Total Tax", value=f"${total_taxed:,}", inline=False)
        locations = "\n ".join(list(locations))
        e.description = f'Locations tracked so far ({len(list(locations))})\n\n {locations}'

        await ctx.channel.send(embed=e)

    @pinger_commands.command(name='inactive', guild_ids=get_all_servers())
    async def inactive_moons(self, ctx, own_corp: Optional[bool] = False):
        """
        Print inactive Moons!
        """
        if own_corp:
            if not self.sender_has_corp_moon_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)
            user = get_auth_user(ctx.user, ctx.guild).profile.main_character
            corps = CorporationAudit.objects.filter(
                corporation__corporation_id=user.corporation_id)
            corp_names = [f"{c.corporation.corporation_name}" for c in corps]
            if corps.count() > 0:
                await ctx.respond(f"Printing Inactive Drills for {', '.join(corp_names)}", ephemeral=True)
                await ctx.author.send(f"Printing Inactive Drills for {', '.join(corp_names)}")
            else:
                await ctx.respond(f"Your corp is not setup in Audit, please contact an admin.", ephemeral=True)

        else:
            if not self.sender_has_moon_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)
            corps = CorporationAudit.objects.filter(
                corporation__corporation_id__in=app_settings.PUBLIC_MOON_CORPS)
            corp_names = [f"{c.corporation.corporation_name}" for c in corps]
            if corps.count() > 0:
                await ctx.respond(f"Printing Inactive Drills for {', '.join(corp_names)}")
            else:
                await ctx.respond(f"Public corp is not setup, please contact an admin.")

        tzactive = timezone.now()
        fracks = MoonFrack.objects.filter(
            arrival_time__gte=tzactive, corporation__in=corps).values_list('structure_id')
        structures = Structure.objects.filter(structureservice__name__in=[
            "Moon Drilling"], corporation__in=corps).exclude(structure_id__in=fracks)
        messages = [f"{s.name}" for s in structures]
        n = 10
        chunks = [list(messages[i * n:(i + 1) * n])
                  for i in range((len(messages) + n - 1) // n)]
        for chunk in chunks:
            message = "\n".join(chunk)
            if own_corp:
                await ctx.author.send(f"```{message}```")
            else:
                await ctx.send(f"```{message}```")

    if app_settings.MOONS_ENABLE_RENT_COG:
        rental_commands = SlashCommandGroup(
            "moon_rentals", "Moon Rental Commands", guild_ids=get_all_servers())

        async def search_moons(ctx: AutocompleteContext):
            """Returns a list of moons that begin with the characters entered so far."""
            resp = list(MapSystemMoon.objects.filter(
                name__icontains=ctx.value).values_list("name", flat=True)[:10])
            return resp

        async def search_characters(ctx: AutocompleteContext):
            """Returns a list of colors that begin with the characters entered so far."""
            resp = list(EveCharacter.objects.filter(
                character_name__icontains=ctx.value).values_list('character_name', flat=True)[:10])
            return resp

        async def search_corp(ctx: AutocompleteContext):
            """Returns a list of colors that begin with the characters entered so far."""
            resp = list(EveCorporationInfo.objects.filter(
                corporation_name__icontains=ctx.value).values_list('corporation_name', flat=True)[:10])
            return resp

        @rental_commands.command(name='status', guild_ids=get_all_servers())
        @option("moon", description="Search for a Moon!", autocomplete=search_moons)
        async def moon_rental_status(self, ctx, moon: str):
            """
            Print Moons Status!
            """
            ctx.defer()
            if not self.sender_has_moon_rental_create_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)

            moon_q = MoonRental.objects.filter(
                moon__name=moon, end_date__isnull=True)

            msgs = []
            for m in MoonRental.objects.filter(moon__name=moon):
                msgs.append(
                    f"{m.start_date.strftime('%y-%m-%d')} to {m.end_date.strftime('%y-%m-%d') if m.end_date else ' ACTIVE '} by {m.contact} [{m.corporation}] for ${m.price:,}")
            msgs = "\n".join(msgs)

            if not moon_q.exists():
                return await ctx.respond(f"{moon} Available!\n```\n{msgs}\n```")
            else:
                return await ctx.respond(f"{moon} is rented!\n```\n{msgs}\n```")

        @rental_commands.command(name='character_status', guild_ids=get_all_servers())
        @option("character", description="Search for a Character!", autocomplete=search_characters)
        async def moon_rental_character_status(self, ctx, character: str):
            """
            Print Moons Status!
            """
            ctx.defer()
            if not self.sender_has_moon_rental_create_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)

            moon_q = MoonRental.objects.filter(
                contact__character_name=character, end_date__isnull=True)

            msgs = []
            for m in moon_q:
                msgs.append(
                    f"{m.moon} by {m.contact} [{m.corporation}] for ${m.price:,}")
            msgs = "\n".join(msgs)

            if not moon_q.exists():
                return await ctx.respond(f"{character} has no rentals.")
            else:
                return await ctx.respond(f"{character} has rented!\n```\n{msgs}\n```")

        @rental_commands.command(name='rent', guild_ids=get_all_servers())
        @option("moon", description="Search for a Moon!", autocomplete=search_moons)
        @option("character", description="Search for a Character!", autocomplete=search_characters)
        @option("corporation", description="Search for a Corporation!", autocomplete=search_corp)
        @option("price", description="Price per month!")
        async def moon_rental_rent(self, ctx, moon: str, character: str, corporation: str, price: int = 100000000):
            """
            Rent a moon!
            """
            ctx.defer()
            if not self.sender_has_moon_rental_create_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)

            moon_q = MoonRental.objects.filter(
                moon__name=moon, end_date__isnull=True)

            if not moon_q.exists():
                moon = MapSystemMoon.objects.get(name=moon)
                char = EveCharacter.objects.get(character_name=character)
                corp = EveCorporationInfo.objects.get(
                    corporation_name=corporation)
                MoonRental.objects.create(moon=moon, contact=char, corporation=corp,
                                          price=price, start_date=timezone.now(), note=f"rented by {ctx.author}")
                return await ctx.respond(f"Rented `{moon}` to `{character} [{corporation}]` for ${price:,}")
            else:
                return await ctx.respond(f"**Unable to rent** `{moon}` to `{character}` Already rented!")

        @rental_commands.command(name='unrent', guild_ids=get_all_servers())
        @option("moon", description="Search for a Moon!", autocomplete=search_moons)
        @option("character", description="Search for a Character!", autocomplete=search_characters)
        async def moon_rental_unrent(self, ctx, moon: str, character: str):
            """
            Rent a moon!
            """
            ctx.defer()
            if not self.sender_has_moon_rental_create_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)

            moon_q = MoonRental.objects.filter(
                moon__name=moon, end_date__isnull=True)

            if moon_q.exists():
                m = moon_q.first()
                m.end_date = timezone.now()
                m.note += f"\nUnrented by {character}, completed by {ctx.author}"
                m.save()
                return await ctx.respond(f"Unrented `{moon}` from `{m.contact}` by `{character}`")
            else:
                return await ctx.respond(f"**Unable to unrent** `{moon}` no rental found?")


def setup(bot):
    bot.add_cog(MoonsCog(bot))



from django.core.management.base import BaseCommand

from moons.tasks import update_ore_prices


class Command(BaseCommand):
    help = 'Update Ore Price Data'

    def handle(self, *args, **options):
        self.stdout.write("Updating Ore Prices!")
        print(update_ore_prices())
        self.stdout.write("Complete!")
